import json
from pathlib import Path
from typing import NamedTuple

Key = str


def is_hash(path: Path):
    return path.name.endswith('.hash')


def to_hash(path: Path):
    assert not is_hash(path)
    return path.with_name(f'{path.name}.hash')


def from_hash(path: Path):
    assert is_hash(path)
    return path.with_name(path.stem)


class FileHash(NamedTuple):
    key: Key
    path: Path
    hash: Path


class TreeHash(NamedTuple):
    key: Key
    path: Path
    hash: Path


class InsideTreeHash(NamedTuple):
    key: Key
    root: Path
    hash: Path
    relative: Path


def dispatch_hash(path):
    path = Path(path)
    key = load_key(path)
    stripped = strip_tree(key)
    if key == stripped:
        return FileHash(key, from_hash(path), path)
    return TreeHash(stripped, from_hash(path), path)


def load_key(path: Path):
    with open(path, 'r') as file:
        return file.read().strip()


def load_tree(path: Path):
    with open(path, 'r') as file:
        return json.load(file)


# FIXME: the names are misleading
def load_tree_key(path: Path):
    with open(path) as file:
        return strip_tree(file.read().strip())


def strip_tree(key):
    # TODO: remove this
    if key.startswith('tree:'):
        key = key[5:]
    elif key.startswith('T:'):
        key = key[2:]
    return key



from pathlib import Path
from datetime import datetime

from tqdm import tqdm

from ..hash import from_hash, is_hash
from ..shortcuts import get_current_repo
from ..utils import call


def blame(path: Path, relative: str):
    repo = get_current_repo()
    path = Path(path).resolve().relative_to(repo.root)
    relative = str(relative)

    folder = path
    if is_hash(folder):
        folder = from_hash(folder)
    base = repo.get_key(folder, relative, version=repo.current_version)

    idx = 0
    bar = tqdm()
    while True:
        idx += 1
        output = call(f'git log -n 1 --skip {idx} --pretty=format:%H,%ct -- {path}', repo.root)
        if not output:
            break

        commit, time = output.split(',', 1)
        bar.update()
        bar.set_description_str(str(datetime.fromtimestamp(int(time))))

        current = repo.load_tree(path, commit)
        if relative not in current or current[relative] != base:
            bar.close()
            print(call(f"git log --format='%an <%ae> at %aD' {commit}^!", repo.root))
            return

    bar.close()
    print(call(f"git log -n 1 --format='%an <%ae> at %aD' -- {path}", repo.root))



from typing import Sequence
from pathlib import Path

from ..interface import Repository
from ..shortcuts import get_consistent_repo
from ..hash import is_hash, to_hash, load_tree, load_key, strip_tree
from ..utils import HashNotFound


def _fetch(repo: Repository, path: Path):
    if not is_hash(path):
        path = to_hash(path)

    key = load_key(path)
    if key.startswith(('T:', 'tree:')):
        key = strip_tree(key)
        keys = repo.storage.read(load_tree, key).values()
    else:
        keys = [key]

    missing = repo.storage.fetch(keys, verbose=True)
    if missing:
        raise HashNotFound(f'Could not fetch {len(missing)} keys from remote.')


def fetch(paths: Sequence[str], context: str = '.'):
    repo = get_consistent_repo([context, *paths])
    for path in paths:
        _fetch(repo, Path(context) / path)



import shutil

from pathlib import Path

from tqdm import tqdm

from ..shortcuts import get_current_repo
from ..hash import is_hash, to_hash, from_hash, load_tree, dispatch_hash, TreeHash
from .add import save_tree


def pull(source: str, destination: str, mode: str):
    repo = get_current_repo()

    assert mode in PULL_MODES, mode
    source, destination = Path(source), Path(destination)
    if not is_hash(source):
        source = to_hash(source)

    if not source.exists():
        raise FileNotFoundError(source)

    h = dispatch_hash(source)
    if isinstance(h, TreeHash):
        mapping = repo.storage.read(load_tree, h.key)

        for file, value in tqdm(mapping.items()):
            file = destination / file
            file.parent.mkdir(parents=True, exist_ok=True)
            PULL_MODES[mode](value, file, repo)

    else:
        PULL_MODES[mode](h.key, destination, repo)


def save_hash(value, file, repo):
    with open(to_hash(file), 'w') as f:
        f.write(value)


PULL_MODES = {
    'copy': lambda h, dst, repo: repo.storage.read(shutil.copyfile, h, dst),
    'hash': save_hash,
}


def gather(source: str, destination: str):
    repo = get_current_repo()

    source, destination = Path(source), Path(destination)
    if not is_hash(destination):
        destination = to_hash(destination)

    if not source.exists():
        raise FileNotFoundError(source)

    tree = {}
    files = [file for file in source.glob('**/*') if not file.is_dir()]
    for file in tqdm(files):
        assert is_hash(file)
        relative = file.relative_to(source)
        with open(file, 'r') as f:
            tree[str(from_hash(relative))] = f.read()

    save_tree(repo, tree, destination)



import shutil
from pathlib import Path

from ..shortcuts import get_consistent_repo
from ..hash import is_hash, load_tree, load_tree_key
from .add import add_folder, save_tree


def update(source: str, destination: str, keep: bool, overwrite: bool):
    source, destination = Path(source), Path(destination)
    if not is_hash(destination):
        raise ValueError('The destination must be a hash.')
    if not source.is_dir():
        raise ValueError('The source must be a folder.')

    repo = get_consistent_repo(['.', source, destination.parent])
    key = load_tree_key(destination)
    mapping = repo.storage.read(load_tree, key)

    new = add_folder(repo, source, None, keep=True)
    common = set(new) & set(mapping)
    mismatch = {k for k in common if new[k] != mapping[k]}
    if mismatch:
        if not overwrite:
            raise ValueError(f'Mismatch between old and new files at following paths: {mismatch}')
        print(f'Mismatched {len(mismatch)} files. Overwriting.')
    if common:
        print(f'{len(common)} files were already present.')

    mapping.update(new)
    save_tree(repo, mapping, destination)
    if not keep:
        shutil.rmtree(source)



from typing import Sequence

from .config import CONFIG, find_repo_root
from .interface import Repository
from .utils import InconsistentRepositories, PathLike, RepositoryNotFound


def get_current_repo(path: PathLike = '.') -> Repository:
    root = find_repo_root(path)
    if root is None:
        raise RepositoryNotFound(f'{CONFIG} files not found in current folder\'s parents')

    return Repository(root)


def get_consistent_repo(paths: Sequence[PathLike]) -> Repository:
    roots = set(filter(None, map(find_repo_root, paths)))
    if len(roots) > 1:
        raise InconsistentRepositories('The paths are located in different repositories')
    if not roots:
        raise RepositoryNotFound(f'{CONFIG} files not found among folder\'s parents')

    return Repository(roots.pop())



from argparse import ArgumentParser

from .add import add
from .blame import blame
from .fetch import fetch
from .init import init
from .pull import pull, gather, PULL_MODES
from .storage import init_storage
from .update import update


def entrypoint():
    parser = ArgumentParser()
    subparsers = parser.add_subparsers()

    new = subparsers.add_parser('init')
    new.set_defaults(callback=init)
    new.add_argument('repository', default='.', nargs='?')
    new.add_argument('-p', '--permissions')
    new.add_argument('-g', '--group')

    new = subparsers.add_parser('add')
    new.set_defaults(callback=add)
    new.add_argument('source')
    new.add_argument('destination')
    new.add_argument('-k', '--keep', default=False, action='store_true')

    new = subparsers.add_parser('fetch')
    new.set_defaults(callback=fetch)
    new.add_argument('paths', nargs='+')

    new = subparsers.add_parser('pull')
    new.set_defaults(callback=pull)
    new.add_argument('source')
    new.add_argument('destination')
    new.add_argument('--mode', choices=list(PULL_MODES), help='how to pull the files from the hash.')

    new = subparsers.add_parser('gather')
    new.set_defaults(callback=gather)
    new.add_argument('source')
    new.add_argument('destination')

    new = subparsers.add_parser('update')
    new.set_defaults(callback=update)
    new.add_argument('source')
    new.add_argument('destination')
    new.add_argument('-k', '--keep', default=False, action='store_true')
    new.add_argument('--overwrite', default=False, action='store_true')

    new = subparsers.add_parser('blame')
    new.set_defaults(callback=blame)
    new.add_argument('path')
    new.add_argument('relative')

    # add_storage_functions(subparsers.add_parser('storage'))

    args = vars(parser.parse_args())
    if 'callback' not in args:
        parser.print_help()
    else:
        callback = args.pop('callback')
        callback(**args)


def add_storage_functions(parser: ArgumentParser):
    subparsers = parser.add_subparsers()
    new = subparsers.add_parser('init')
    new.set_defaults(callback=init_storage)



from typing import Union

import yaml

from bev.config import find_repo_root, CONFIG, load_config
from bev.utils import RepositoryNotFound
from connectome.storage.config import DiskConfig, STORAGE_CONFIG_NAME, root_params
from connectome.storage.utils import mkdir


def init(repository: str = '.', permissions: str = None, group: Union[int, str] = None):
    root = find_repo_root(repository)
    if root is None:
        raise RepositoryNotFound(f'{CONFIG} files not found in current folder\'s parents')

    return init_config(load_config(root / CONFIG), permissions, group)


def init_config(config, permissions, group):
    local, meta = config.local, config.meta
    digest_size = meta.hash.build()().digest_size
    permissions, group = get_root_params(local.storage + local.cache, permissions, group)

    for location in local.storage + local.cache:
        storage_root = location.root
        if not storage_root.exists():
            mkdir(storage_root, permissions, group, parents=True)

        conf_path = storage_root / STORAGE_CONFIG_NAME
        if not conf_path.exists():
            with open(conf_path, 'w') as file:
                yaml.safe_dump(DiskConfig(
                    hash=meta.hash, levels=[1, digest_size - 1]
                ).dict(exclude_defaults=True), file)


def get_root_params(entries, permissions, group):
    for entry in entries:
        if entry.root.exists():
            return root_params(entry.root)

    if permissions is None:
        permissions = input('Folder permissions:')
    if isinstance(permissions, str):
        permissions = int(permissions, base=8)
    assert 0 <= permissions <= 0o777
    if group is None:
        group = input('Folder group:')
    return permissions, group



import inspect
import os
import subprocess
import warnings
from functools import lru_cache
from pathlib import Path
from typing import Sequence, Union

from wcmatch import glob

from .config import CONFIG, build_storage, find_vcs_root
from .hash import is_hash, to_hash, load_tree, load_tree_key, strip_tree, FileHash, InsideTreeHash, Key
from .local import Local, LocalVersion
from .utils import InconsistentRepositories, call_git, HashNotFound, PathLike, RepositoryNotFound

CommittedVersion = str
Version = Union[CommittedVersion, LocalVersion]


class Repository:
    """
    Interface that represents a `bev` repository.

    Parameters
    ----------
    root:
        the path to the repository's root, i.e. the folder that contains `.bev.config`
    fetch: bool
        whether to fetch files from remote locations when needed. Can be overridden in corresponding methods
    version: str, Local
        default value for data version. Can be either a string with a commit hash/tag or the `Local` object.
        Can be overridden in corresponding methods
    """

    def __init__(self, *root: PathLike, fetch: bool = True, version: Version = None):
        self.root = Path(*root)
        self.storage, self.cache = build_storage(self.root)
        self._fetch, self._version = fetch, version

    @classmethod
    def from_root(cls, *parts: PathLike):
        warnings.warn(
            'This classmethod is deprecated, call the constructor directly for the same behaviour', UserWarning)
        return cls(*parts)

    @classmethod
    def from_here(cls, *relative: PathLike, fetch: bool = True, version: Version = None) -> 'Repository':
        """
        Creates a repository with a path `relative` to the file in which this method is called.

        Examples
        --------
        >>> repo = Repository.from_here('../../data')
        """
        file = Path(inspect.stack()[1].filename)
        return cls(file.parent / Path(*relative), fetch=fetch, version=version)

    @classmethod
    def from_vcs(cls, *parts: PathLike) -> 'Repository':
        vcs = find_vcs_root(Path(os.getcwd(), *parts))
        if vcs is None:
            raise RepositoryNotFound(f'{Path(*parts)} is not inside a vcs repository')

        configs = list(vcs.rglob(CONFIG))
        if not configs:
            raise RepositoryNotFound(f'{Path(*parts)} is not inside a bev repository')
        if len(configs) > 1:
            raise InconsistentRepositories(f'This vcs repository contains multiple bev repositories: {configs}')

        return cls(configs[0].parent)

    @property
    def current_version(self) -> CommittedVersion:
        return self.latest_version()

    def latest_version(self, path: PathLike = '.') -> CommittedVersion:
        path = Path(path)
        if not (self.root / path).exists() and not is_hash(path):
            path = to_hash(path)

        if not (self.root / path).exists():
            raise FileNotFoundError(path)

        return call_git(f'git log -n 1 --pretty=format:%H -- {path}', self.root)

    def resolve(self, *parts: PathLike, version: Version = None, fetch: bool = None) -> Path:
        version = self._resolve_version(version)
        key = self.get_key(*parts, version=version, fetch=fetch)
        return self.storage.resolve(key, fetch=self._resolve_fetch(fetch))

    def glob(self, *parts: PathLike, version: Version = None, fetch: bool = None) -> Sequence[Path]:
        version = self._resolve_version(version)
        h = self._split(Path(*parts), version)
        if not isinstance(h, InsideTreeHash):
            raise ValueError('`glob` is only applicable to tree hashes')

        pattern = str(h.relative)
        tree = self._load(load_tree, h.key, fetch)
        files = sorted(self._expand_folders(tree))

        return [h.root / file for file in glob.globfilter(files, pattern, flags=glob.GLOBSTAR)]

    # TODO: cache this based on path parents
    def get_key(self, *parts: PathLike, version: Version = None, fetch: bool = None) -> Key:
        version = self._resolve_version(version)
        path = Path(*parts)
        h = self._split(path, version)
        if isinstance(h, FileHash):
            return h.key

        assert isinstance(h, InsideTreeHash), h
        relative = str(h.relative)
        tree = self._get_tree(h.key, version, fetch)
        if relative not in tree:
            if relative in self._expand_folders(tree):
                raise HashNotFound(f'"{str(path)}" is a folder inside a tree hash')

            raise HashNotFound(str(path))

        return tree[relative]

    def load_tree(self, path: PathLike, version: Version = None, fetch: bool = None) -> dict:
        version = self._resolve_version(version)
        key = self._get_hash(Path(path), version)
        if key is None:
            raise HashNotFound(path)
        return self._get_tree(key, version, fetch)

    # internal logic

    def _get_tree(self, key, version, fetch):
        # we need the version here, because we want to cache only a committed tree
        if version == Local:
            return self._load(load_tree, key, fetch)
        return self._load_cached_tree(key, fetch)

    @lru_cache(None)
    def _load_cached_tree(self, key, fetch):
        return self._load(load_tree, key, fetch=fetch)

    def _get_hash(self, path: Path, version: Version):
        if version == Local:
            return self._get_uncomitted_hash(path)
        return self._get_committed_hash(path, version)

    def _get_uncomitted_hash(self, relative: Path):
        path = self.root / relative
        if path.exists():
            return load_tree_key(path)

    @lru_cache(None)
    def _get_committed_hash(self, relative: Path, version: str):
        assert isinstance(version, str), type(version)
        relative = str(relative)
        if not relative.startswith('./'):
            relative = f'./{relative}'

        try:
            return strip_tree(call_git(f'git show {version}:{relative}', self.root))
        except subprocess.CalledProcessError:
            pass

    def _load(self, func, key, fetch):
        fetch = self._resolve_fetch(fetch)
        return self.storage.read(func, key, fetch=fetch)

    def _split(self, path: Path, version: Version):
        # TODO: use bin-search?
        for parent in list(reversed(path.parents))[1:]:
            hash_path = to_hash(parent)
            key = self._get_hash(hash_path, version)
            if key is not None:
                key = strip_tree(key)
                return InsideTreeHash(key, parent, hash_path, path.relative_to(parent))

        hash_path = to_hash(path)
        key = self._get_hash(hash_path, version)
        if key is None:
            raise HashNotFound(path)

        return FileHash(key, path, hash_path)

    def _resolve_fetch(self, fetch):
        if fetch is None:
            return self._fetch
        return fetch

    def _resolve_version(self, version):
        if version is None:
            version = self._version
        if version is None:
            raise ValueError('The argument `version` must be provided')
        return version

    @staticmethod
    def _expand_folders(tree) -> set:
        result = set(tree)
        for file in tree:
            result.update(map(str, list(Path(file).parents)[:-1]))
        return result



import ast
import hashlib
import inspect
from pathlib import Path

from connectome.storage.config import init_storage as _init_storage


def init_storage():
    path = Path()
    if list(path.iterdir()):
        print('The folder is not empty')
        exit(1)

    algos = hashlib.algorithms_guaranteed
    default_algo = 'sha256'

    print(f'Hash algorithm (Press Enter for {default_algo}, ? for options): ', end='')
    option = input().strip() or default_algo
    while option not in algos:
        if option != '?':
            print(f'Unknown algorithm "{option}"')
        print('Available algorithms:', ', '.join(algos))
        print('Hash algorithm (Press Enter for sha256, ? for options): ', end='')
        option = input().strip() or default_algo

    algo = getattr(hashlib, option)
    try:
        params = {x.name: x.default for x in list(inspect.signature(algo).parameters.values())[1:]}
    except ValueError:
        params = {}

    kwargs = {}
    if params:
        all_params = '\n'.join(f'{k}: {v}' for k, v in kwargs.items())
        print(f'Default hash parameters:\n{all_params}\nModify hash parameters? (y/N): ', end='')
        if parse_yes():
            for param, default in params.items():
                print(f'Algorithm parameter "{param}" (Press Enter for {default}): ', end='')
                value = input().strip()
                if value:
                    value = ast.literal_eval(value)
                    if value != default:
                        kwargs[param] = value

    digest_size = algo(**kwargs).digest_size
    half_digest = digest_size // 2
    default_levels = 1, half_digest - 1, half_digest
    print(f'Folder levels. Must sum to {digest_size} (Press Enter for {default_levels}): ', end='')
    levels = parse_levels(default_levels, digest_size)

    assert 'name' not in kwargs
    kwargs['name'] = option
    _init_storage(path, algorithm=kwargs, levels=levels, exist_ok=True)


def parse_yes():
    raw = input().strip()
    value = raw.lower()
    while True:
        if value in ['y', 'yes']:
            return True
        if value in ['n', 'no', '']:
            return False

        print('Unknown option:', raw, 'Try again: ', end='')
        raw = input().strip()
        value = raw.lower()


def parse_levels(default, size):
    levels = input().strip()
    while True:
        if not levels:
            return default

        levels = ast.literal_eval(levels)
        if sum(levels) == size:
            return levels

        print(f"The levels don't sum to {size}. Try again: ", end='')
        levels = input().strip()



from pathlib import Path
import subprocess
import shlex
from typing import Union

PathLike = Union[str, Path]


# TODO: use gitpython
def call_git(command: str, cwd=None) -> str:
    return subprocess.check_output(shlex.split(command), cwd=cwd, stderr=subprocess.DEVNULL).decode('utf-8').strip()


def call(command: str, cwd=None) -> str:
    try:
        return subprocess.check_output(shlex.split(command), cwd=cwd).decode('utf-8').strip()
    except subprocess.CalledProcessError as e:
        raise RuntimeError(e.stderr or e.stdout) from e


class HashNotFound(Exception):
    pass


class RepositoryError(Exception):
    pass


class RepositoryNotFound(RepositoryError):
    pass


class InconsistentRepositories(RepositoryError):
    pass


# legacy
HashNotFoundError = HashNotFound



import json
import shutil
import os
from collections import OrderedDict
from pathlib import Path
from typing import Optional

from tqdm import tqdm

from ..interface import Repository, PathLike
from ..shortcuts import get_consistent_repo
from ..hash import is_hash, to_hash


def validate_file(path: Path):
    assert path.is_file()
    if is_hash(path):
        raise ValueError('You are trying to add a hash to the storage.')

    return path


def save_tree(repo: Repository, tree: dict, destination: Path):
    # save the directory description
    # making sure that each time the same string will be saved
    tree = OrderedDict((k, tree[k]) for k in sorted(tree))
    tree_path = destination.parent / f'{destination.name}.hash.temp'

    # TODO: storage should allow writing directly from memory
    with open(tree_path, 'w') as file:
        json.dump(tree, file)
    key = repo.storage.write(tree_path)
    os.remove(tree_path)

    with open(destination, 'w') as file:
        file.write(f'T:{key}')

    return key


def add_file(repo: Repository, source: Path, destination: Optional[Path], keep: bool):
    validate_file(source)
    key = repo.storage.write(source)

    if destination is not None:
        assert is_hash(destination)
        with open(destination, 'w') as file:
            file.write(key)

    if not keep:
        os.remove(source)

    return key


def add_folder(repo: Repository, source: Path, destination: Optional[Path], keep: bool):
    assert source.is_dir()

    tree = {}
    files = [file for file in source.glob('**/*') if not file.is_dir()]
    for file in tqdm(files):
        relative = file.relative_to(source)
        tree[str(relative)] = add_file(repo, file, None, True)

    result = tree
    if destination is not None:
        result = save_tree(repo, tree, destination)
    if not keep:
        shutil.rmtree(source)

    return result


def add(source: PathLike, destination: PathLike, keep: bool, context: str = '.'):
    source, destination = Path(source), Path(destination)

    if not source.exists():
        raise FileNotFoundError(source)

    if destination.is_dir():
        destination /= to_hash(source).name
    if not is_hash(destination):
        destination = to_hash(destination)
    if destination.exists():
        raise FileExistsError(destination)

    if source.is_dir():
        repo = get_consistent_repo([context, destination.parent, source])
        add_folder(repo, source, destination, keep)
    else:
        repo = get_consistent_repo([context, destination.parent, source.parent])
        add_file(repo, source, destination, keep)



class LocalVersion:
    """A class used to denote the current uncommitted version of data"""

    def __getstate__(self):
        raise RuntimeError('This object cannot be pickled')

    def __eq__(self, other):
        return isinstance(other, LocalVersion)


Local = LocalVersion()



from pathlib import Path

from setuptools import setup, find_packages

classifiers = '''Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Programming Language :: Python :: 3.8
Programming Language :: Python :: 3.9'''

with open('README.md', encoding='utf-8') as file:
    long_description = file.read()

with open('requirements.txt', encoding='utf-8') as file:
    requirements = file.read().splitlines()

# get the current version
with open(Path(__file__).resolve().parent / 'bev/__version__.py', encoding='utf-8') as file:
    scope = {}
    exec(file.read(), scope)
    __version__ = scope['__version__']

setup(
    name='bev',
    packages=find_packages(include=('bev',)),
    include_package_data=True,
    version=__version__,
    description='A small manager for versioned data',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT',
    url='https://github.com/neuro-ml/bev',
    download_url='https://github.com/neuro-ml/bev/archive/v%s.tar.gz' % __version__,
    keywords=['data', 'version control'],
    classifiers=classifiers.splitlines(),
    install_requires=requirements,
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'bev = bev.cli.entrypoint:entrypoint',
        ],
    },
)



import os
import re
import socket
import warnings
from itertools import chain
from pathlib import Path
from typing import Dict, Sequence, Tuple, Callable, Any, Union, NamedTuple
import importlib

from paramiko.config import SSHConfig
from pydantic import BaseModel, Extra, validator, root_validator
from yaml import safe_load
from connectome.storage import Storage, SSHLocation, Disk, RemoteStorage
from connectome.storage.config import HashConfig

from .utils import PathLike

CONFIG = '.bev.yml'


class ConfigError(Exception):
    pass


class HostName:
    key: str

    def match(self, name) -> bool:
        raise NotImplementedError

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if isinstance(v, HostName):
            return v
        if isinstance(v, str):
            return StrHostName(v)

        assert isinstance(v, dict), f'Not a dict: {v}'
        assert len(v) == 1
        (k, v), = v.items()
        for kls in HostName.__subclasses__():
            if kls.key == k:
                return kls(v)

        raise ValueError(f'Invalid key "{k}" for hostname')


class StrHostName(HostName):
    key = 'str'

    def __init__(self, value):
        self.value = value

    def match(self, name) -> bool:
        return name == self.value


class RegexHostName(HostName):
    key = 'regex'

    def __init__(self, pattern):
        self.pattern = re.compile(pattern)

    def match(self, name) -> bool:
        return self.pattern.match(name) is not None


class LocationConfig(BaseModel):
    root: Path
    ssh: str = None

    @root_validator(pre=True)
    def resolve_deprecation(cls, values):
        if 'host' in values:
            assert 'ssh' not in values
            warnings.warn('The name "host" has been renamed to "ssh"')
            values['ssh'] = values.pop('host')

        return values

    class Config:
        extra = Extra.forbid


LocationConfigs = Tuple[LocationConfig, ...]


class StorageConfig(BaseModel):
    name: str
    default: Dict[str, Any] = None
    hostname: Tuple[HostName, ...] = ()
    storage: LocationConfigs
    cache: LocationConfigs = ()

    @validator('hostname', pre=True)
    def from_single(cls, v):
        if isinstance(v, (str, dict)):
            v = v,
        return v

    @validator('storage', 'cache', pre=True)
    def from_string(cls, v):
        if isinstance(v, str):
            v = v,
        return v

    @validator('storage', 'cache', each_item=True, pre=True)
    def add_defaults(cls, v, values):
        default = (values['default'] or {}).copy()
        if isinstance(v, str):
            v = {'root': v}
        assert isinstance(v, dict)
        default.update(v)
        return default

    class Config:
        extra = Extra.forbid


class ConfigMeta(BaseModel):
    choose: str = None
    fallback: str = None
    order: str = None
    hash: Union[str, HashConfig] = None

    @validator('hash', pre=True)
    def normalize_hash(cls, v):
        if isinstance(v, str):
            v = {'name': v}
        return v

    @root_validator(pre=True)
    def resolve_deprecation(cls, values):
        if 'default' in values:
            assert 'fallback' not in values
            warnings.warn('The name "default" has been renamed to "fallback"')
            values['fallback'] = values.pop('default')

        return values

    class Config:
        extra = Extra.forbid


class RepositoryConfig(BaseModel):
    local: StorageConfig
    remotes: Tuple[StorageConfig, ...]
    meta: ConfigMeta

    class Config:
        extra = Extra.forbid


class CacheIndex(NamedTuple):
    local: Sequence[PathLike]
    remote: Sequence[RemoteStorage] = ()


def load_config(config: Path) -> RepositoryConfig:
    with open(config, 'r') as file:
        return parse(config, safe_load(file))


def is_remove_available(location, config):
    # TODO: better way of handling missing hosts
    return location.ssh is not None and (config.lookup(location.ssh) != {
        'hostname': location.ssh} or location.ssh in config.get_hostnames())


def build_storage(root: Path) -> Tuple[Storage, CacheIndex]:
    config = load_config(root / CONFIG)
    meta = config.meta

    order_func: Callable[[Sequence[Disk]], Sequence[Disk]] = identity
    if meta.order is not None:
        path, attr = meta.order.rsplit('.', 1)
        order_func = getattr(importlib.import_module(path), attr)

    remote_storage, remote_cache = [], []
    # filter only available hosts
    # TODO: move to config?
    config_path = Path('~/.ssh/config').expanduser()
    if config_path.exists():
        with open(config_path) as f:
            ssh_config = SSHConfig()
            ssh_config.parse(f)

            for entry in config.remotes:
                for location in entry.storage:
                    if is_remove_available(location, ssh_config):
                        remote_storage.append(SSHLocation(location.ssh, location.root))

                for location in entry.cache:
                    if is_remove_available(location, ssh_config):
                        remote_cache.append(SSHLocation(location.ssh, location.root))

    loc = order_func([Disk(location.root) for location in config.local.storage])
    return Storage(loc, remote_storage), CacheIndex([c.root for c in config.local.cache], remote_cache)


def parse(root, config) -> RepositoryConfig:
    if not isinstance(config, dict):
        raise ConfigError('The config must be a dict')

    meta = ConfigMeta.parse_obj(config.pop('meta', {}))
    entries = {}
    for name, entry in config.items():
        if isinstance(entry, str):
            entry = {'storage': entry}
        if not isinstance(entry, dict):
            raise ConfigError('Each config entry must be either a dict or a string')
        if 'name' in entry:
            raise ConfigError('The key "name" is not available')
        entry = entry.copy()
        entry['name'] = name
        entries[name] = StorageConfig(**entry)

    fallback = None
    filter_func: Callable[[StorageConfig], bool] = default_choose
    if meta.choose is not None:
        path, attr = meta.choose.rsplit('.', 1)
        filter_func = getattr(importlib.import_module(path), attr)
    if meta.fallback is not None:
        fallback = meta.fallback
        if fallback not in entries:
            raise ConfigError(f'The fallback ({fallback}) is not present in the config {root}')

    if len(entries) == 1:
        local, = entries.values()
        remotes = []
    else:
        name = choose_local(entries.values(), filter_func, fallback)
        if name is None:
            raise ConfigError(f'No matching entry in config {root}')
        local = entries.pop(name)
        remotes = list(entries.values())

    return RepositoryConfig(local=local, remotes=remotes, meta=meta)


def choose_local(metas, func, default):
    for meta in metas:
        if func(meta):
            return meta.name

    return default


def default_choose(meta: StorageConfig):
    repo_key = 'BEV__REPOSITORY'
    if repo_key in os.environ:
        return meta.name == os.environ[repo_key]

    node = socket.gethostname()
    hosts = meta.hostname or [StrHostName(meta.name)]
    return any(h.match(node) for h in hosts)


def identity(x):
    return x


def _find_root(path, marker):
    path = Path(path).resolve()
    for parent in chain([path], path.parents):
        if (parent / marker).exists():
            return parent


def find_repo_root(path: PathLike):
    return _find_root(path, CONFIG)


def find_vcs_root(path: PathLike):
    return _find_root(path, '.git')
